#! /usr/bin/env python

import logging

def get_logger():

    logger = logging.getLogger("chimera_utils")
    logger.propagate = False
 
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
     
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s', datefmt='%m/%d/%Y %H:%M:%S')

    ch.setFormatter(formatter)

    logger.addHandler(ch)

    return logger




#! /usr/bin/env python

from .parser import create_parser

def main():

    cparser = create_parser()
    args = cparser.parse_args()

    if vars(args) == {}:
        cparser.print_usage()
    else:
        args.func(args)




#! /usr/bin/env python

from .run import *

import argparse
from .version import __version__

def create_parser():

    parser = argparse.ArgumentParser(prog = "chimera_utils")
    parser.add_argument("--version", action = "version", version = "%(prog)s " + __version__)

    subparsers = parser.add_subparsers()

    ##########
    # imple intron retention count 

    count = subparsers.add_parser("count",
                                  help = "Count supporting read pairs for each chimeric junction")

    count.add_argument("chimeric_sam", metavar = "Chimeric.out.sam", default = None, type = str,
                              help = "the path to the chimeric read file generated at STAR alignment step")

    count.add_argument("output_file", metavar = "output_file", default = None, type = str, 
                              help = "the path to the output")

    count.add_argument("--abnormal_insert_size", type = int, default = 500000,
                       help = "size of abnormal insert size. used for checking the consistency of paired read of breakpoint containing reads (default: %(default)s)")

    count.add_argument("--min_major_clipping_size", type = int, default = 15,
                       help = "minimum number of clipped bases for junction read (default: %(default)s)")

    count.add_argument("--min_read_pair_num", type = int, default = 3,
                       help = "minimum required number of supporting junction read pairs (default: %(default)s)")

    count.add_argument("--min_cover_size", type = int, default = 30,
                       help = "region size which have to be covered by aligned short reads (default: %(default)s)")

    count.add_argument("--anchor_size_thres", type = int, default = 10,
                       help = "at least an anchor size of one chimeric read have to be equal or larger than the specified value (default: %(default)s)")
     
    count.add_argument("--min_chimeric_size", type = int, default = 1000,
                       help = "threshold of minimum chimeric transcript sizes (default: %(default)s)")

    count.add_argument("--debug", default = False, action = 'store_true', help = "keep intermediate files")

    count.set_defaults(func = count_main)
    ##########

    # merge control 
    merge_control = subparsers.add_parser("merge_control",
                                          help = "Merge, compress and index chimeric count files generated by count command")

    merge_control.add_argument("chimeric_count_list", metavar = "chimeric_count_list.txt", default = None, type = str,
                               help = "chimeric count file list")

    merge_control.add_argument("output_file", default = None, type = str,
                               help = "the path of the output file")

    merge_control.set_defaults(func = merge_control_main)

    """
    ##########
    # filter

    filter = subparsers.add_parser("filter",
                                   help = "filter out intron retentions that do not satisty specified conditions")

    filter.add_argument("intron_retention_file", metavar = "intron_retention.txt", default = None, type = str,
                        help = "the path to intron retention file generated by simple_count command")

    filter.add_argument("output_file", metavar = "output.txt", default = None, type = str,
                        help = "the path to the output file")

    filter.add_argument("--num_thres", type = int, default = 3,
                        help = "remove intron retentions whose supporting read numbers are below this value (default: %(default)s)")

    filter.add_argument("--ratio_thres", type = int, default = 0.05,
                        help = "remove intron retentions whose whose ratios (Intron_Retention_Read_Count / Edge_Read_Count) \
                        are below this value (default: %(default)s)")
                               
    filter.add_argument("--pooled_control_file", default = None, type = str,
                        help = "the path to control data created by merge_control (default: %(default)s)")

    filter.set_defaults(func = filter_main)
    """
    ##########
    # associate

    associate = subparsers.add_parser("associate", 
                                      help = "Associate chimeric junctions with the output of GenomonSV")

    associate.add_argument("chimera_file", metavar = "chimera.txt", default = None, type = str,
                           help = "the path to chimera file")

    associate.add_argument("genomonSV_file", metavar = "genomonSV.result.txt", default = None, type = str,
                           help = "the path to GenomonSV result file")

    associate.add_argument("output_file", metavar = "output.txt", default = None, type = str,
                             help = "the path to the output")

    associate.add_argument("--margin", default = 10, type = int,
                              help = "the margin for comparing gene fusions")

    associate.add_argument("--sv_margin_major", default = 500000, type = int,
                              help = "the margin for comparing gene fusions and SVs")

    associate.add_argument("--sv_margin_minor", default = 10, type = int,
                              help = "the margin for comparing gene fusions and SVs")

    associate.add_argument("--genome_id", choices = ["hg19", "hg38", "mm10"], default = "hg19",
                           help = "the genome id used for selecting UCSC-GRC chromosome name corresponding files (default: %(default)s)")

    associate.add_argument("--grc", default = False, action = 'store_true',
                           help = "convert chromosome names to Genome Reference Consortium nomenclature (default: %(default)s)")

    associate.add_argument("--debug", default = False, action = 'store_true', help = "keep intermediate files")

    associate.set_defaults(func = associate_main)
    ##########

    return parser





#! /usr/bin/env python

from __future__ import print_function

from fusionfusion.config import *
from fusionfusion import region_utils
from fusionfusion import regions

def get_chimera_info(input_file, output_file):

    min_read_pair_num = param_conf.min_read_pair_num
    min_valid_read_pair_ratio = param_conf.min_valid_read_pair_ratio
    min_cover_size = param_conf.min_cover_size
    min_chimeric_size = param_conf.min_chimeric_size
    anchor_size_thres = param_conf.anchor_size_thres


    hin = open(input_file, 'r')
    hout = open(output_file, 'w')

    print('\t'.join(["Chr_1", "Pos_1", "Dir_1", "Chr_2", "Pos_2", "Dir_2", 
                     "Inserted_Seq", "Read_Pair_Num", "Max_Over_Hang_1", "Max_Over_Hang_2"]), file = hout)

    for line in hin:
        F = line.rstrip('\n').split('\t')

        if F[0] == F[3] and abs(int(F[1]) - int(F[4])) < min_chimeric_size: continue

        coveredRegion_primary = F[9].split(';')
        coveredRegion_pair = F[12].split(';')
        coveredRegion_SA = F[15].split(';')

        # check the number of unique read pairs
        coveredRegion_meta = [coveredRegion_primary[i] + ';' + coveredRegion_pair[i] + ';' + coveredRegion_SA[i] for i in range(0, len(coveredRegion_primary))]
        uniqueCoverdRegion_meta = list(set(coveredRegion_meta))
        if len(uniqueCoverdRegion_meta) < min_read_pair_num: continue

        # check the maximum anchor size
        coverRegionSize_primary = list(map(region_utils.getCoverSize, coveredRegion_primary))
        coverRegionSize_SA = list(map(region_utils.getCoverSize, coveredRegion_SA))
        anchor_size = [min(coverRegionSize_primary[i], coverRegionSize_SA[i]) for i in range(len(coverRegionSize_primary))]
        if max(anchor_size) < anchor_size_thres: continue

        # check for the covered region
        pairPos = F[17].split(';')
        primaryPos = F[18].split(';')

        region1 =  regions.Regions()
        region2 =  regions.Regions()

        for i in range(0, len(coveredRegion_primary)):
            if primaryPos[i] == "1" and pairPos != "0":
                for reg in coveredRegion_primary[i].split(','):
                    region1.addMerge(reg)
            elif primaryPos[i] == "2" and pairPos != "0":
                for reg in coveredRegion_primary[i].split(','):
                    region2.addMerge(reg)

        for i in range(0, len(coveredRegion_SA)):
            if primaryPos[i] == "1" and pairPos != "0":
                for reg in coveredRegion_SA[i].split(','):
                    region2.addMerge(reg)
            elif primaryPos[i] == "2" and pairPos != "0":
                for reg in coveredRegion_SA[i].split(','):
                    region1.addMerge(reg)

        for i in range(0, len(coveredRegion_pair)):
            if (primaryPos[i] == "1" and pairPos[i] == "1") or (primaryPos[i] == "2" and pairPos[i] == "2"):
                for reg in coveredRegion_pair[i].split(','):
                    region1.addMerge(reg)
            elif (primaryPos[i] == "1" and pairPos[i] == "2") or (primaryPos[i] == "2" and pairPos[i] == "1"):
                for reg in coveredRegion_pair[i].split(','):
                    region2.addMerge(reg)

        region1.reduceMerge()
        region2.reduceMerge()

        if region1.regionSize() < min_cover_size or region2.regionSize() < min_cover_size: continue


        print('\t'.join(F[0:7]) + '\t' + str(len(uniqueCoverdRegion_meta)) + '\t' + str(region1.regionSize()) + '\t' + str(region2.regionSize()), file = hout)

    hin.close()
    hout.close()




#! /usr/bin/env python

from __future__ import print_function
import re


def convert_to_bedpe(input_file, output_file, margin_major, margin_minor):

    hin = open(input_file, 'r')
    hout = open(output_file, 'w')
    for line in hin:
        F = line.rstrip('\n').split('\t')
        if F[0].startswith('#'): continue
        if F[0] == "Chr_1" and F[1] == "Pos_1" and F[2] == "Dir_1": continue

        chr1, pos1, dir1, chr2, pos2, dir2, inseq = F[0], F[1], F[2], F[3], F[4], F[5], F[6]
        start1, end1, start2, end2 = pos1, pos1, pos2, pos2
        ID = ','.join([chr1, pos1, dir1, chr2, pos2, dir2, inseq])

        read_num = 0

        if dir1 == '+':
            start1 = str(int(start1) - int(margin_minor))
            end1 = str(int(end1) + int(margin_major))
        else:
            start1 = str(int(start1) - int(margin_major))
            end1 = str(int(end1) + int(margin_minor))

        if dir2 == '+':
            start2 = str(int(start2) - int(margin_minor))
            end2 = str(int(end2) + int(margin_major))
        else:
            start2 = str(int(start2) - int(margin_major))
            end2 = str(int(end2) + int(margin_minor))

        print('\t'.join([chr1, start1, end1, chr2, start2, end2, ID, str(read_num), dir1, dir2]), file = hout)

    hin.close()
    hout.close()





#! /usr/bin/env python

import pysam

junction_margin = 5

def get_gene_info(chr, pos, gene_tb):

    tabixErrorFlag = 0
    try:
        records = gene_tb.fetch(chr, int(pos) - 1, int(pos) + 1)
    except Exception as inst:
        # print >> sys.stderr, "%s: %s" % (type(inst), inst.args)
        tabixErrorFlag = 1
        
    gene = [];
    if tabixErrorFlag == 0:
        for record_line in records:
            record = record_line.split('\t')
            gene.append(record[3])

    return sorted(list(set(gene)))


def get_junc_info(chr, pos, exon_tb, junction_margin):

    tabixErrorFlag = 0
    try:
        records = exon_tb.fetch(chr, int(pos) - junction_margin, int(pos) + junction_margin)
    except Exception as inst:
        # print >> sys.stderr, "%s: %s" % (type(inst), inst.args)
        tabixErrorFlag = 1
        
    junction = []
    if tabixErrorFlag == 0:
        for record_line in records:
            record = record_line.split('\t')
            if abs(int(pos) - int(record[1])) < junction_margin:
                if record[5] == "+": junction.append(record[3] + ".start")
                if record[5] == "-": junction.append(record[3] + ".end")
            if abs(int(pos) - int(record[2])) < junction_margin:
                if record[5] == "+": junction.append(record[3] + ".end")
                if record[5] == "-": junction.append(record[3] + ".start")

    return sorted(list(set(junction)))





#!/usr/bin/env python


from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

def get_version():
    with open(path.join(here, "chimera_utils/version.py"), encoding = 'utf-8') as hin:
        for line in hin:
            if line.startswith("__version__"):
                version = line.partition('=')[2]
                return version.strip().strip('\'"')
    raise ValueError('Could not find version.')

setup(
    name = 'chimera_utils',
    version = get_version(),
    description='Python tools for processing chimeric reads and lists of gene fusions.',
    long_description=long_description, 
    long_description_content_type='text/markdown',
    url = 'https://github.com/friend1ws/chimera_utils',
    author = 'Yuichi Shiraishi',
    author_email = 'friend1ws@gamil.com',
    license = 'GPLv3',

    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: Unix',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],

    packages = find_packages(exclude = ['tests']),

    install_requires = ["annot_utils>=0.3.0", "fusionfusion>=0.5.0", "pysam>=0.9.0"],
    entry_points = {'console_scripts': ['chimera_utils = chimera_utils:main']}

)




#! /usr/bin/env python

from __future__ import print_function

import sys, os, subprocess
import pysam

import fusionfusion.parseJunctionInfo
from fusionfusion.config import *

from . import count, process
from .logger import get_logger
logger = get_logger()

def count_main(args):

    param_conf.debug = args.debug
    param_conf.abnormal_insert_size = args.abnormal_insert_size
    param_conf.min_major_clipping_size = args.min_major_clipping_size
    param_conf.min_read_pair_num = args.min_read_pair_num
    param_conf.min_cover_size = args.min_cover_size
    param_conf.anchor_size_thres = args.anchor_size_thres
    param_conf.min_chimeric_size = args.min_chimeric_size


    fusionfusion.parseJunctionInfo.parseJuncInfo_STAR(args.chimeric_sam, args.output_file + ".chimeric.tmp.txt")

    hout = open(args.output_file + ".chimeric.txt", 'w')
    subprocess.check_call(["sort", "-f", "-k1,1", "-k2,2n", "-k4,4", "-k5,5n", args.output_file + ".chimeric.tmp.txt"], stdout = hout)
    hout.close()

    fusionfusion.parseJunctionInfo.clusterJuncInfo(args.output_file + ".chimeric.txt",
                                                   args.output_file + ".chimeric.clustered.txt")

    count.get_chimera_info(args.output_file + ".chimeric.clustered.txt", args.output_file)

    if param_conf.debug == False:
        subprocess.check_call(["rm", "-rf", args.output_file + ".chimeric.tmp.txt"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".chimeric.txt"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".chimeric.clustered.txt"])


def merge_control_main(args):

    # make directory for output if necessary
    if os.path.dirname(args.output_file) != "" and not os.path.exists(os.path.dirname(args.output_file)):
        os.makedirs(os.path.dirname(args.output_file))

    subprocess.check_call(["rm", "-rf", args.output_file + ".unsorted"])
    subprocess.check_call(["touch", args.output_file + ".unsorted"])
    hout = open(args.output_file + ".unsorted", 'a')

    with open(args.chimeric_count_list, 'r') as hin:
        for line in hin:
            count_file = line.rstrip('\n')
            with open(count_file, 'r') as hin2:
                for line2 in hin2:
                    if line2.startswith("Chr_1"): continue
                    F = line2.rstrip('\n').split('\t')
                    print('\t'.join(F[:7]), file = hout)
            
            # the following code produces errors... need to investigate subprocess.. 
            # tail = subprocess.Popen(["tail", "-n", "+2", count_file], stdout = subprocess.PIPE)
            # cut = subprocess.Popen(["cut", "-f1-7"], stdin = tail.stdout, stdout = hout)

    hout.close()


    hout = open(args.output_file + ".sorted", 'w')
    s_ret = subprocess.check_call(["sort", "-f", "-k1,1", "-k2,2n", "-k4,4", "-k5,5n", args.output_file + ".unsorted"], stdout = hout)
    hout.close()

    hout = open(args.output_file + ".merged", 'w')
    s_ret = subprocess.check_call(["uniq", args.output_file + ".sorted"], stdout = hout)
    hout.close()

    hout = open(args.output_file, 'w')
    s_ret = subprocess.check_call(["bgzip", "-f", "-c", args.output_file + ".merged"], stdout = hout)
    hout.close()

    if s_ret != 0:
        print("Error in compression merged junction file", file = sys.stderr)
        sys.exit(1)


    s_ret = subprocess.check_call(["tabix", "-p", "vcf", args.output_file])
    if s_ret != 0:
        print("Error in indexing merged junction file", file = sys.stderr)
        sys.exit(1)

    subprocess.check_call(["rm", "-f", args.output_file + ".unsorted"])
    subprocess.check_call(["rm", "-f", args.output_file + ".sorted"])
    subprocess.check_call(["rm", "-f", args.output_file + ".merged"])


def associate_main(args): 

    from annot_utils.utils import grc_check

    if args.grc == True:
        logger.warning("--grc argument is deprecated and ignored.")

    is_grc = grc_check(args.chimera_file, [0, 3])
    is_grc_sv = grc_check(args.genomonSV_file, [0, 3])

    if is_grc != is_grc_sv:
        logger.warning("Chimeric count file and SV file seems to use different coordinate system.")

    process.convert_to_bedpe(args.chimera_file, args.output_file + ".fusion.bedpe", args.sv_margin_major, args.sv_margin_minor)
    process.convert_to_bedpe(args.genomonSV_file, args.output_file + ".genomonSV.bedpe", args.margin, args.margin)

    hout = open(args.output_file + ".fusion_comp.bedpe", 'w')
    subprocess.check_call(["bedtools", "pairtopair", "-a", args.output_file + ".fusion.bedpe", "-b", args.output_file + ".genomonSV.bedpe"], stdout = hout)
    hout.close()

    # create dictionary
    chimera2sv = {}
    hin = open(args.output_file + ".fusion_comp.bedpe", 'r')
    for line in hin:
        F = line.rstrip('\n').split('\t')
        chimera2sv[F[6]] = F[16]
    
    hin.close()

    from annot_utils.gene import make_gene_info  
    from annot_utils.exon import make_exon_info
    from .annotation import get_gene_info, get_junc_info

    make_gene_info(args.output_file + ".refGene.bed.gz", "refseq", args.genome_id, is_grc, True)
    make_exon_info(args.output_file + ".refExon.bed.gz", "refseq", args.genome_id, is_grc, True)
    make_gene_info(args.output_file + ".ensGene.bed.gz", "gencode", args.genome_id, is_grc, False)
    make_exon_info(args.output_file + ".ensExon.bed.gz", "gencode", args.genome_id, is_grc, False)

    ref_gene_tb = pysam.TabixFile(args.output_file + ".refGene.bed.gz")
    ref_exon_tb = pysam.TabixFile(args.output_file + ".refExon.bed.gz")
    ens_gene_tb = pysam.TabixFile(args.output_file + ".ensGene.bed.gz")
    ens_exon_tb = pysam.TabixFile(args.output_file + ".ensExon.bed.gz")


    # add SV annotation to fusion
    hout = open(args.output_file, 'w')
    with open(args.chimera_file, 'r') as hin:
        header = hin.readline().rstrip('\n')
        print(header + '\t' + '\t'.join(["Gene_1", "Gene_2", "Junc_1", "Junc_2", "Chimera_Class", "SV_Key"]), file = hout)
 
        for line in hin:
            F = line.rstrip('\n').split('\t')
            ID = ','.join([F[0], F[1], F[2], F[3], F[4], F[5], F[6]])

            if ID not in chimera2sv: continue
            SV_info = chimera2sv[ID]

            ref_gene_info_1 = get_gene_info(F[0], F[1], ref_gene_tb) 
            ens_gene_info_1 = get_gene_info(F[0], F[1], ens_gene_tb)
            gene_info_1 = ref_gene_info_1 if len(ref_gene_info_1) > 0 else ens_gene_info_1

            ref_gene_info_2 = get_gene_info(F[3], F[4], ref_gene_tb)
            ens_gene_info_2 = get_gene_info(F[3], F[4], ens_gene_tb)
            gene_info_2 = ref_gene_info_2 if len(ref_gene_info_2) > 0 else ens_gene_info_2
        
            junc_info_1 = get_junc_info(F[0], F[1], ref_exon_tb, 5) 
            if len(junc_info_1) == 0: junc_info_1 = get_junc_info(F[0], F[1], ens_exon_tb, 5)
        
            junc_info_2 = get_junc_info(F[3], F[4], ref_exon_tb, 5) 
            if len(junc_info_2) == 0: junc_info_2 = get_junc_info(F[3], F[4], ens_exon_tb, 5)

            same_gene_flag = False        
            for g1 in ref_gene_info_1 + ens_gene_info_1:
                for g2 in ref_gene_info_2 + ens_gene_info_2:
                    if g1 == g2: same_gene_flag = True

            gene_info_str_1 = "---" if len(gene_info_1) == 0 else ';'.join(sorted(list(set(gene_info_1))))
            gene_info_str_2 = "---" if len(gene_info_2) == 0 else ';'.join(sorted(list(set(gene_info_2))))
            junc_info_str_1 = "---" if len(junc_info_1) == 0 else ';'.join(sorted(list(set(junc_info_1))))
            junc_info_str_2 = "---" if len(junc_info_2) == 0 else ';'.join(sorted(list(set(junc_info_2))))


            sv_chr1, sv_pos1, sv_dir1, sv_chr2, sv_pos2, sv_dir2, sv_inseq = SV_info.split(',')

        
            if ("start" in junc_info_str_1 and "end" in junc_info_str_2) or ("start" in junc_info_str_2 and "end" in junc_info_str_1):
                if sv_dir1 == '-' and sv_dir2 == '+' and same_gene_flag == True:
                    chimera_type = "Exon reusage"
                else:
                    chimera_type = "Spliced chimera"
            elif abs(int(F[1]) - int(sv_pos1)) < 10 and abs(int(F[4]) - int(sv_pos2)) < 10:
                chimera_type = "Unspliced chimera"
            else:
                chimera_type = "Putative spliced chimera"


            print('\t'.join(F) + '\t' + gene_info_str_1 + '\t' + gene_info_str_2 + '\t' + \
                  junc_info_str_1 + '\t' + junc_info_str_2 + '\t' + \
                  chimera_type + '\t' + SV_info, file = hout)

    hin.close()
    hout.close()


    if args.debug == False:
        subprocess.check_call(["rm", "-rf", args.output_file + ".fusion.bedpe"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".genomonSV.bedpe"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".fusion_comp.bedpe"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".refGene.bed.gz"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".refGene.bed.gz.tbi"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".refExon.bed.gz"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".refExon.bed.gz.tbi"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".ensGene.bed.gz"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".ensGene.bed.gz.tbi"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".ensExon.bed.gz"])
        subprocess.check_call(["rm", "-rf", args.output_file + ".ensExon.bed.gz.tbi"])
  



